
/*
   Myfv.c

  This program allows to display binary files one screenful at a time or
  a certain number of lines, as selected by the user, and prompts the user
  for extra display.  It allows to go back in the file, forward, to print line
  numbers and to search for strings.

  This code has been modified from the solution to assignment 3, written by
  Mashrur Mia.

  Marc-Andre Laverdiere
  ID:4469526
  Department of Computer Science
  Concordia University
  (c) 2001

  Note: this program can be modified and distributed, as long as the credentials are kept integrally in the distribution.
*/

/*
  myfv_dyn.c
  This program, when linked with myfv_main.c and myfv_static.c, will produce the program myfv.
 This code is to be compiled as a dynamic library.  It is entirely standalone.
*/


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>  
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define BUFSIZE 1024
#define DEFAULT 22
#define SCREENWIDTH 75


int cfp = 0,nread=0,fileEOF = 0;
int lastLine = 0;
int lineshown = 0;
int buffernum = 0;
char buf[BUFSIZE];


void error(char *str){
  perror(str);
  _exit(0);
}


int occurs(char *line, char *str)

/* returns 0 if the str occurs in line; -1 otherwise */

{
   int len = strlen(str);
   int len2 = strlen(line)-len;
   char *s1;
   
   s1 = line;
   for ( ; s1<=line+len2; s1++) 
       if (strncmp(s1,str,len) == 0)
           return 0;
   return -1;
}



void reset(){
  cfp=nread=0;
  fileEOF=0;
  buffernum = 0;
}


int readline(int cfd, char *line, int n){
  
  int i=0;

  do{
    if(cfp>=nread){         /* when buffer is needed to be loaded */
      nread = 0;            /* saving few conditions*/
      if(!fileEOF){         /* file already has reached EOF */
	if((nread = read(cfd,buf,BUFSIZE))<0)
	  error("read");
	if(nread!=BUFSIZE)  /* file EOF has reached*/
	  fileEOF = 1;
	if(nread){          /* something in buffer */
	  buffernum++;
	  cfp = 0;
	}
      }
    }
  }while(nread && (!((line[i++]=buf[cfp++])=='\n' || i >= n-1)));
  
  line[i]='\0';    
  
  return i;
}


void show_next(int file, int lines){
  int i=0;
  int n;
  int j =0;
  int TooBigFinalString = 0;
  char buf[256];
  for(; i<lines && (n=readline(file,buf,256))>0 ;i++,j++){
    /* controls strings larger than the width of the screen*/
    if (n > SCREENWIDTH)
      {
	i += n/SCREENWIDTH;
	/*	if (i>= DEFAULT)
	  {
	    lseek(file,-n+1,SEEK_CUR);
	    j--;
	    TooBigFinalString = -1;
	  }
      }    
    if (TooBigFinalString)
      break;*/
      }
    write(1,buf,n);
    if(buf[n-1]!='\n')
      write(1,"\n",1);
  }
  lineshown = lineshown + j;
  if(j){
    if (i<lines)
      for(;j<lines;j++)
        write(1,"\n",1);
    else{
    lastLine=j==lines?DEFAULT:(DEFAULT-lines+j);
    for(;j<=lastLine;j++)
      write(1,"\n",1);
    }
  }
}

int readback(int fd,int nline){
  int thisBufSize=0;
  if(cfp<2 && buffernum<=1){
    cfp = 0;
    return 0;
  }
  cfp--;
  do{
    if(cfp < 1 && nline){
      if (buffernum>1){
	thisBufSize = lseek(fd,0,SEEK_CUR);
	thisBufSize = thisBufSize % BUFSIZE;
	thisBufSize = thisBufSize?thisBufSize:BUFSIZE;
	lseek(fd,-(BUFSIZE+thisBufSize),SEEK_CUR);
	
	nread = read(fd,buf,BUFSIZE);
	buffernum--;
	fileEOF=0;
	cfp = nread;
      }	
      else
	nline=0;
    }
    if(nline && buf[cfp-1]=='\n')
      nline--;
  }while(nline && cfp--);
  return 1;
}

void show_prev(int file, int lines){
  int lineback = 0;
  
  if(lineshown>DEFAULT && lines>0){
    lineback = lines+lastLine;
    if(readback(file,lineback)){
      lineshown = lineshown - lineback;
      show_next(file,DEFAULT);
    }
  }
}



void add_line_nos(int srcfile, int dstfile){
  char lineRead[256];
  char lineWrite[300];
  int lineNum = 0;
  reset();
  while(readline(srcfile,lineRead,256)){
    sprintf(lineWrite,"%d\t%s",++lineNum,lineRead);
    write(dstfile,lineWrite,strlen(lineWrite));
  }
}

void get_str_lines(int srcfile, int dstfile, char *str){
  char lineRead[256];
  reset();
  while(readline(srcfile,lineRead,256))
    if(occurs(lineRead,str)==0)
      write(dstfile,lineRead,strlen(lineRead));
}

void get_str_lines_nos(int srcfile, int dstfile, char *str){
  char lineRead[256];
  char lineWrite[300];

  int lineNum = 0;
  reset();
  while(readline(srcfile,lineRead,256)){
    lineNum++;
    if(occurs(lineRead,str)==0){
      sprintf(lineWrite,"%d\t%s",lineNum,lineRead);
      write(dstfile,lineWrite,strlen(lineWrite));
    }
  }
}

void skip_lines(int file, int lines)
{
  int i=0;
  int n;
  long TotalSeek = 0;
  char Message[256];
  char buf[256];
  sprintf(Message,"\nFile has less than %d lines.\n",lines);
  reset();
  for(; i<lines && (n=readline(file,buf,256))>=0 ;i++)
  {
    if (n <= 0)
      {
	write(2,Message,strlen(Message));
	exit(1);
      }
    TotalSeek += n;
  }
  lseek(file,TotalSeek,SEEK_SET);
}



